#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>

#include <signal.h>
#include <setjmp.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>

#define MIRROR	"kbe.rotabanner.com"
#define LOGIN	"login"
#define OPTIONS	""

#define TIMEOUT 2
#define CHARSET getenv("SOURCE_CHARSET")

#include "time.h"

/* ************************ string functions *********************** */

inline char *nl(char *str) { return str ? str:""; }
inline char *newstr(char *str) { return str ? strdup(str):(char *)NULL; };
inline void freestr(char *str) { if (str) free(str); };

void putline(char **buf,char *str){
    if (str) if (*buf){
	int ofs=strlen(*buf),
	    len=strlen(str)+1;
	*buf=(char *)realloc(*buf,ofs+len);
	memcpy(&(*buf)[ofs],str,len);
    } else *buf=newstr(str);
}

char *getline(char **buf){
    char *result=*buf;
    if (result) {
	*buf=strchr(result,'\n');
	if (*buf) *((*buf)++)='\0'; else if (!*result) return NULL;
    }
    return result;
}

char *strjoin(char *first, ...){
    char *result=newstr(first),*item;
    va_list iterator;
    va_start(iterator,first);
    while (item=va_arg(iterator,char *)) putline(&result,item);
    va_end(iterator);
    return result;
}

/* ************************ http stuff *********************** */

int sock;

int set_addr(struct sockaddr_in *addr,char *host,short port){
    memset(addr,0,sizeof(*addr));
    addr->sin_family=AF_INET;
    addr->sin_port=htons(port);
    if (host){
	struct hostent *hostIP;
	if (!(hostIP=gethostbyname(host))) return 0;
	memcpy(&addr->sin_addr,hostIP->h_addr,hostIP->h_length);
    } else addr->sin_addr.s_addr=INADDR_ANY;
    return 1;
}

#define BSIZE 4096
char *read_reply(int sock){
    int len=0;
    char *buf=NULL;

    while (buf=(char *)realloc(buf,len+BSIZE)){
	int rlen=read(sock,&buf[len],BSIZE);
	if (!rlen) break;
	len+=rlen;
    }
    if (len){
	buf=(char *)realloc(buf,len+1);
	buf[len]='\0';
    } else {
	freestr(buf);
	buf=NULL;
    }
    return buf;
}

char *http(char *host,char *request){
    struct sockaddr_in dest;
    return host && request && set_addr(&dest,host,80) && (sock=socket(AF_INET,SOCK_STREAM,0))>=0 && 
	   !connect(sock,(struct sockaddr *)&dest,sizeof(dest)) && write(sock,request,strlen(request))==strlen(request) ?
	   read_reply(sock):(char *)NULL;
}

/* ********************** request *************************** */

char *extract_banner(char *buf){
    int enter=0;
    if (strstr(nl(getline(&buf))," 200") && buf) for(;*buf;buf++)
	if (*buf=='\n') {
	    if (enter) return newstr(buf+1); else enter=1;
	} else if (*buf!='\r') enter=0;
    return NULL;
}

static sigjmp_buf timeout_env;
void timeout_handler(int sig){
    siglongjmp(timeout_env,1);
}

char *request(char *login,char *mirror,char *query,int timeout,char *headers,char *random){
    char *result,*banner,
	 *request=strjoin("GET /cgi-bin/iframe",login,"?",random,"&ssi=yes",query," HTTP/1.0\nHOST: ",mirror,"\n",nl(headers),"\n",NULL);
    sock=-1;
    signal(SIGALRM,timeout_handler);
    alarm(timeout);
    result=sigsetjmp(timeout_env,1) ? (char *)NULL : http(mirror,request);
    if (sock>=0) close(sock);

    banner=extract_banner(result);
    freestr(result);
    freestr(request);

    return banner;
}

char *default_banner(char *login,char *mirror,char *query,char *random){
    return strjoin("<a href=\"http://",mirror,"/cgi-bin/href",login,"?",random,"\" target=_top><img src=\"http://",mirror,
           "/cgi-bin/banner",login,"?",random,query,"\" alt=\"List Banner Exchange\" width=470 height=60 border=0 ismap></a>",NULL);
}

char *banner(char *login,char *query){
    int i;
    char *result,*headers=NULL;
    char random[32];
    struct {
	char *name,*value;
    } header[]={
	{"REFERER",strjoin("http://",nl(getenv("SERVER_NAME")),getenv("REQUEST_URI"),NULL)},
	{"USER-AGENT",getenv("HTTP_USER_AGENT")},
	{"X_FORWARDED_FOR",getenv("HTTP_X_FORWARDED_FOR")},
	{"REMOTE_ADDR",getenv("REMOTE_ADDR")},
    };
    for(i=0;i<sizeof(header)/sizeof(header[0]);i++)
	if (header[i].value) {
	    putline(&headers,header[i].name);
	    putline(&headers,": ");
	    putline(&headers,header[i].value);
	    putline(&headers,"\n");
	}

    srand(getpid()+time(NULL));
    snprintf(random,32,"%d",rand()%10000000);
    query=strjoin("&charset=",nl(CHARSET),(query && *query ? "&":""),nl(query),NULL);

    result=request(login,MIRROR,query,TIMEOUT,headers,random);
    if (!result) result=default_banner(login,MIRROR,query,random);

    freestr(query);
    freestr(headers);
    freestr(header[0].value);
    
    return result;
}

/* ********************** main function *************************** */
int main(){
    char *query=getenv("QUERY_STRING"),
         *login=getenv("PATH_INFO"),
         *data=banner(login && *login ? login:"/"LOGIN, query && *query ? query:OPTIONS);
    int length=strlen(nl(data));
    printf("Content-type: text/html\nContent-Length: %d\n\n",length);
    if (data && length){
	fwrite(data,1,length,stdout);
	free(data);
    }
}
